import numpy as np
from scipy.signal import lfilter


class Processor:
    """Base class for processors

    A processor is an object that manipulates the data in some way.
    """
    def __init__(self, device=None, **kwargs):
        self.device = device
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __call__(self, frame):
        return self.process(frame)

    def process(self, frame):
        """Processes a single fraame of input.

        The input frame might be the rame object as the read frame, so a
        processor should not manitulate the data in place.

        Arguments:
            frame (`numpy.ndarray`): ``(n_ch, n_samp)`` shape input frame.
        """

    def setup(self):
        pass

    def reset(self):
        pass

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        self._device = device


class LevelDetector(Processor):
    """Single channel level detector.

    A level detector that tracks the root-mean-square level in a signal.
    The level is tracked with an exponentially decaying time average,
    implemented as a low-passed squared amplitude.

    Arguments:
        channel (`int`): The index of the channel to track.
        time_constant (`float`): Time constant for the exponential decay in
            seconds, default 50 ms.
        **kwargs: Extra arguments will be passed to `Processor`.
    """
    def __init__(self, channel, time_constant=50e-3, **kwargs):
        Processor.__init__(self, **kwargs)
        self.time_constant = time_constant

        self.channel = channel
        self.reset()

    def process(self, frame):
        # Squared input level is tracked in order for the RMS trigger to work properly.
        input_levels = frame[self.channel]**2
        output_levels, self._buffer = lfilter([self._digital_constant], [1, self._digital_constant - 1], input_levels, zi=self._buffer)
        return output_levels**0.5

    def reset(self):
        self._buffer = np.atleast_1d(0)

    @property
    def time_constant(self):
        return -1 / (np.log(1 - self._digital_constant) * self.device.fs)

    @time_constant.setter
    def time_constant(self, val):
        self._digital_constant = 1 - np.exp(-1 / (val * self.device.fs))

    @property
    def current_level(self):
        return self._buffer**0.5


class LevelDetectorAttackRelease:
    def __init__(self, *, channel, fs, time_constant=None, attack_time=None, release_time=None):
        self.fs = fs
        if attack_time is not None and release_time is not None:
            # separate attack and release times specified, used them
            # Attack and release time are defined as the time it takes for a step
            # to rise from 10% to 90%.
            self.attack_time = attack_time
            self.release_time = release_time
        elif time_constant is not None:
            # Single number or two numbers
            self.time_constant = time_constant
        else:
            self.time_constant = 50e-3  # Default value of 50 ms

        self.channel = channel
        self.current_level = 0

    def __call__(self, block):
        """
        Todo:
            This is too slow. If this style of detector is kept it must be improved
            with better processing. Possibilities are cython, numbe, or faust.
        """
        input_levels = np.abs(block[self.channel])
        output_levels = np.empty_like(input_levels)
        level = self.current_level  # Get the level after the privious block
        for idx in range(len(input_levels)):
            if input_levels[idx] > level:
                level = self._attack_constant * input_levels[idx] + (1 - self._attack_constant) * level
            else:
                level = self._release_constant * input_levels[idx] + (1 - self._release_constant) * level
            output_levels[idx] = level
        self.current_level = level  # Save the level to use in the next block
        return output_levels

    @property
    def attack_time(self):
        return -2.2 / self.fs / np.log(1 - self._attack_constant)

    @attack_time.setter
    def attack_time(self, value):
        self._attack_constant = 1 - np.exp(-2.2 / (value * self.fs))

    @property
    def release_time(self):
        return -2.2 / self.fs / np.log(1 - self._release_constant)

    @release_time.setter
    def release_time(self, value):
        self._release_constant = 1 - np.exp(-2.2 / (value * self.fs))

    @property
    def time_constant(self):
        values = (self.attack_time, self.release_time)
        if (values[1] - values[0]) * 2 / (values[1] + values[0]) < 0.01:
            # The relative difference is less than 1%, retuen just a single value
            return values[0]
        else:
            return values

    @time_constant.setter
    def time_constant(self, value):
        try:
            len(value)
        except TypeError:
            value = (value, value)

        # if len(value) is not 2:
        self.attack_time = value[0] * 2.2
        self.release_time = value[1] * 2.2



import numpy as np
import threading
import logging

from . import processors

logger = logging.getLogger(__name__)


class Trigger:
    """Base class for Trigger implementation.

    A Trigger is an object that performs a test on all input data from a
    Device, regardless if the input is set as active or not. If the test
    evaluates to ``True`` the trigger will perform a set of actions,
    e.g. activate the input of a Device.

    Arguments:
        actions (callable or list of callables): The actions that will be
            called each time the test evaluates to ``True``.
        false_actions (callable or list of callables): The actions that will be
            called each time the test evaluates to ``False``.
        auto_deactivate (`bool`): Sets if the trigger deactivates itself when
            the test is ``True``. Useful to only trigger once, dafault ``True``.
        use_calibrations (`bool`): Sets if calibration values from the Device
            should be used for the test, default ``True``.
    Attributes:
        active (`~threading.Event`): Controls if the trigger is active or not.
            A deactivated trigger will still test (e.g. to track levels), but
            not take action. Triggers start of as active unless manually deactivated.
    """
    def __init__(self, action=None, false_action=None, auto_deactivate=True,
                 use_calibrations=True, device=None):
        self.device = device
        # self.active = multiprocessing.Event()
        self.active = threading.Event()
        self.active.set()

        self.actions = []
        self.auto_deactivate = auto_deactivate
        self.use_calibrations = use_calibrations
        if action is not None:
            try:
                self.actions.extend(action)
            except TypeError:
                self.actions.append(action)

        self.false_actions = []
        if false_action is not None:
            try:
                self.false_actions.extend(false_action)
            except TypeError:
                self.false_actions.append(false_action)

    def __call__(self, frame):
        """Manages testing and actions."""
        # We need to perform the test event if the triggering is disabled
        # Some triggers (RMSTrigger) needs to update their state continuously to work as intended
        # If e.g. RMSTrigger cannot update the level with the triggering disabled, it will always
        # start form zero
        test = self.test(frame * self.calibrations)
        if self.active.is_set():
            # logger.debug('Testing in {}'.format(self.__class__.__name__))
            if test:
                [action() for action in self.actions]
            else:
                [action() for action in self.false_actions]

    def test(self, frame):
        """Performs test.

        The trigger conditions should be implemented here.

        Arguments:
            frame (`numpy.ndarray`): The current input frame to test.
        Returns:
            `bool`: ``True`` -> do ``actions``, ``False`` -> do ``false_actions``
        """
        raise NotImplementedError('Required method `test` is not implemented in {}'.format(self.__class__.__name__))

    def reset(self):
        """Resets the trigger state."""
        self.active.set()

    def setup(self):
        """Configures trigger state."""
        if self.use_calibrations:
            calibrations = self.device.calibrations
        else:
            calibrations = np.ones(len(self.device.inputs))
        self.calibrations = calibrations[:, np.newaxis]

    @property
    def auto_deactivate(self):
        return self.active.clear in self.actions

    @auto_deactivate.setter
    def auto_deactivate(self, value):
        if value and not self.auto_deactivate:
            self.actions.insert(0, self.active.clear)
        elif self.auto_deactivate and not value:
            self.actions.remove(self.active.clear)

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, dev):
        self._device = dev


class RMSTrigger(Trigger):
    """RMS level trigger.

    Triggers actions based on a detected root-mean-square level.

    Arguments:
        level (`float`): The level at which to trigger.
        channel (`int`): The index of the channel on which to trigger.
        region (``'Above'`` or ``'Below'``, optional): Defines if the triggering
            happens when the detected level rises above or falls below the set
            level, default ``'Above'``.
        level_detector_args (`dict`, optional): Passed as keyword arguments to
            the internal `~.processors.LevelDetector`.
        **kwargs: Extra keyword arguments passed to `Trigger`.
    """
    def __init__(self, level, channel, region='Above', level_detector_args=None, **kwargs):
        Trigger.__init__(self, **kwargs)
        self.channel = channel
        # self.level_detector = LevelDetector(channel=channel, fs=fs, **kwargs)
        self.region = region
        self.trigger_level = level
        self.level_detector_args = level_detector_args if level_detector_args is not None else {}

    def setup(self):
        Trigger.setup(self)
        self.level_detector = processors.LevelDetector(channel=self.channel, device=self.device, **self.level_detector_args)

    def test(self, frame):
        # logger.debug('Testing in RMS trigger')
        levels = self.level_detector(frame)
        return any(self._sign * levels > self.trigger_level * self._sign)

    def reset(self):
        Trigger.reset(self)
        self.level_detector.reset()

    @property
    def region(self):
        if self._sign == 1:
            return 'Above'
        else:
            return 'Below'

    @region.setter
    def region(self, value):
        if value.lower() == 'above':
            self._sign = 1
        elif value.lower() == 'below':
            self._sign = -1
        else:
            raise ValueError('{} not a valid region for RMS trigger.'.format(value))


class PeakTrigger(Trigger):
    """Peak level trigger.

    Triggers actions based on detected peak level.

    Arguments:
        level (`float`): The level at which to trigger.
        channel (`int`): The index of the channel on which to trigger.
        region (``'Above'`` or ``'Below'``, optional): Defines if the triggering
            happens when the detected level rises above or falls below the set
            level, default ``'Above'``.
        **kwargs: Extra keyword arguments passed to `Trigger`.
    """
    def __init__(self, level, channel, region='Above', **kwargs):
        Trigger.__init__(self, **kwargs)
        self.region = region
        self.trigger_level = level
        self.channel = channel

    def test(self, frame):
        # logger.debug('Testing in Peak triggger')
        levels = np.abs(frame[self.channel])
        return any(self._sign * levels > self.trigger_level * self._sign)

    @property
    def region(self):
        if self._sign == 1:
            return 'Above'
        else:
            return 'Below'

    @region.setter
    def region(self, value):
        if value.lower() == 'above':
            self._sign = 1
        elif value.lower() == 'below':
            self._sign = -1
        else:
            raise ValueError('{} not a valid region for peak trigger.'.format(value))


class DelayedAction:
    """Delays an action.

    When called, an instance of this class will excecute a specified action
    after a set delay. This can be useful to create timed measurements or
    pauses in a longer sequence.

    Arguments:
        action (callable): Any callable action. This can be a callable class,
            a user defined funciton, or a method of another class.
            If several actions are required, create a lambda that calls all
            actions when called.
        time (`float`): The delay time, in seconds.
    """
    def __init__(self, action, time):
        self.action = action
        self.time = time
        # self.timer = Timer(interval=time, function=action)

    def __call__(self):
        timer = threading.Timer(interval=self.time, function=self.action)
        timer.start()
        # self.timer.start()



import numpy as np
from numpy.fft import rfft as fft, irfft as ifft
from scipy.signal import waveforms, max_len_seq
import queue
from . import utils


class Generator:
    """Base class for generator implementations.

    A `Generator` is an object that creates data for output channels in a
    Device. Refer to specific generators for more details.
    """
    def __init__(self, device=None, **kwargs):
        self.device = device
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __call__(self):
        """Manages frame creation"""
        return np.atleast_2d(self.frame())

    def frame(self):
        """Generates a frame of output.

        The generated frame must match the device framesize.
        If the generator creates multiple channels, it should have the shape
        ``(n_ch, framesize)``, otherwise 1d arrays are sufficient.

        Returns:
            `numpy.ndarray`: Generated frame.
        """
        raise NotImplementedError('Required method `frame` is not implemented in {}'.format(self.__class__.__name__))

    def reset(self):
        """Resets the generator."""
        pass

    def setup(self):
        """Configures the generator state."""
        pass

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, dev):
        self._device = dev


class GeneratorStop(Exception):
    """Raised by Generators.

    This exception indicates that the generator have reached some stopping
    criteria, e.g. end of file. Should be caught by the Device to stop output.
    """
    pass


class QGenerator(Generator):
    """Generator using `queue.Queue`.

    Implementation of a `Generator` using a queue.
    Takes data from an input queue and generates frames with the correct
    framesize. The input queue must be filled fast enough otherwise the
    device output is cancelled.

    Attributes:
        Q (`~queue.Queue`): The queue from where data is extracted.
    """
    def __init__(self, **kwargs):
        Generator.__init__(self, **kwargs)
        self.Q = queue.Queue()
        self.buffer = None

    def frame(self):
        gen_frame = []
        samples_left = self.device.framesize
        if self.buffer is not None:
            gen_frame.append(self.buffer[..., :samples_left])
            samples_left -= gen_frame[-1].shape[-1]
        while samples_left > 0:
            try:
                frame = self.Q.get(timeout=self.device._generator_timeout)
            except queue.Empty:
                raise GeneratorStop('Input Q is empty')
            gen_frame.append(frame[..., :samples_left])
            samples_left -= frame.shape[-1]
        if samples_left < 0:
            self.buffer = frame[..., samples_left:]
        else:
            self.buffer = None
        return np.concatenate(gen_frame, axis=-1)

    def reset(self):
        """Clears the input queue."""
        Generator.reset(self)
        utils.flush_Q(self.Q)


class ArbitrarySignalGenerator(Generator):
    """Repeated generation of arbritrary signals.

    Implementation of `Generator` for arbritrary signals.

    Arguments:
        repetitions (`float`): The number of cycles to output before stopping, default `np.inf`.
        **kwargs: Will be saved as ``kwargs`` and accessible in `setup`.
    Keyword Arguments:
        signal (`numpy.ndarray`): One cycle of the signal to output.
    """
    def __init__(self, repetitions=np.inf, **kwargs):
        Generator.__init__(self, **kwargs)
        self.repetitions = repetitions  # Default to continious output
        self.reset()

    def frame(self):
        if self.repetitions_done >= self.repetitions:
            raise GeneratorStop('Finite number of repetitions reached')
        samples_left = self.device.framesize
        gen_frame = []
        while samples_left > 0:
            gen_frame.append(self.signal[..., self.idx:self.idx + samples_left])
            samps_app = gen_frame[-1].shape[-1]
            samples_left -= samps_app
            self.idx += samps_app
            if self.idx >= self.signal.shape[-1]:
                self.repetitions_done += 1
                self.idx = 0
                if self.repetitions_done >= self.repetitions:
                    gen_frame.append(np.zeros(self.signal.shape[:-1] + (samples_left,)))
                    samples_left = 0
        return np.concatenate(gen_frame, axis=-1)

    def reset(self):
        Generator.reset(self)
        self.idx = 0
        self.repetitions_done = 0

    def setup(self):
        """Configures the signal.

        Create the signal manually and pass it while creating the generator
        as the ``signal`` argument.

        It is possible to inherit `ArbitrarySignalGenerator` and override the
        setup method. Create one cycle of the signal and store it in
        ``self.signal``. Access the underlying device as ``self.device``,
        which has important properties, e.g. samplerate ``fs``.
        All keyword arguments passed while creating instances are available
        as ``self.key``.

        Note:
            Call `ArbitrarySignalGenerator.setup(self)` from subclasses.
        """
        Generator.setup(self)


class SweepGenerator(ArbitrarySignalGenerator):
    """Swept sine generator.

    Arguments:
        start_frequency (`float`): Initial frequency of the sweep, in Hz.
        stop_frequency (`float`): Final frequency of the sweep, in Hz.
        duration (`float`): Duration of a single sweep, in seconds.
        repetitions (`float`, optional): The number of repetitions, default `np.inf`.
        method (`str`, optional): Chooses the type of sweep, see
            `~scipy.signal.chirp`, default ``'logarithmic'``.
        bidirectional (`bool`, optional): If the sweep is bidirectional or not,
            default ``False``.
    See Also:
        `ArbitrarySignalGenerator`, `scipy.signal.chirp`
    """
    def __init__(self, start_frequency, stop_frequency, duration,
                 method='logarithmic', bidirectional=False, **kwargs):
        ArbitrarySignalGenerator.__init__(self, **kwargs)
        self.start_frequency = start_frequency
        self.stop_frequency = stop_frequency
        self.duration = duration
        self.method = method
        self.bidirectional = bidirectional

    def setup(self):
        ArbitrarySignalGenerator.setup(self)
        time_vector = np.arange(round(self.duration * self.device.fs)) / self.device.fs
        self.signal = waveforms.chirp(time_vector, self.start_frequency, self.duration, self.stop_frequency, method=self.method, phi=90)
        if self.bidirectional:
            self.signal = np.concatenate([self.signal, self.signal[::-1]])
            self.repetitions /= 2


class MaximumLengthSequenceGenerator(ArbitrarySignalGenerator):
    """Generation of maximum length sequences.

    Arguments:
        order (`int`): The order or the sequence. The total length  is ``2**order - 1``.
        repetitions (`float`, optional): The number of repetitions, default `np.inf`.
    See Also:
        `ArbitrarySignalGenerator`, `scipy.signal.max_len_seq`
    """
    def __init__(self, order, **kwargs):
        ArbitrarySignalGenerator.__init__(self, **kwargs)
        self.order = order

    def setup(self):
        ArbitrarySignalGenerator.setup(self)
        self.sequence, state = max_len_seq(self.order)
        self.signal = 1 - 2 * self.sequence


class FunctionGenerator(Generator):
    """Generates signals from a shape function.

    Implementation of `Generator` for standard funcitons.

    Arguments:
        frequency (`float`): The frequecy of the signal, in Hz.
        amplitude (`float`, optional): The amplitude of the signal, default to 1.
        repetitions (`float`, optional): The number of repetitions, default `np.inf`.
        shape (`str`, optional): Function shape, default ``'sine'``. Currently
            available functions are

                - ``'sine'``: `numpy.sin`
                - ``'sawtooth'``: `scipy.signal.sawtooth`
                - ``'square'``: `scipy.signal.square`

        phase_offset (`float`, optional): Phase offset of the signal in radians, default 0.
        shape_kwargs (`dict`): Keyword arguments for shape function.
    """
    _functions = {
        'sin': waveforms.sin,
        'saw': waveforms.sawtooth,
        'squ': waveforms.square
    }

    def __init__(self, frequency, amplitude=1, repetitions=np.inf,
                 shape='sine', phase_offset=0, shape_kwargs=None, **kwargs):
        Generator.__init__(self, **kwargs)
        self.repetitions = repetitions  # Default to continious output
        self.frequency = frequency
        self.amplitude = amplitude
        self.shape = shape
        self.phase_offset = phase_offset
        self.shape_kwargs = {} if shape_kwargs is None else shape_kwargs

    def frame(self):
        if self.repetitions_done >= self.repetitions:
            raise GeneratorStop('Finite number of repetitions reached')
        frame = self._function(self._phase_array + self._phase, **self.shape_kwargs)
        self._phase += self._phase_per_frame
        if self.repetitions_done >= self.repetitions:
            surplus_reps = self.repetitions_done - self.repetitions
            surplus_samps = round(surplus_reps * self.device.fs / self.frequency)
            frame[-surplus_samps:] = 0
        return self.amplitude * frame

    def setup(self):
        Generator.setup(self)
        self.reset()
        taps = np.arange(self.device.framesize)
        self._phase_array = 2 * np.pi * taps * self.frequency / self.device.fs
        self._phase_per_frame = 2 * np.pi * self.frequency / self.device.fs * self.device.framesize

    def reset(self):
        Generator.reset(self)
        self._phase = self.phase_offset

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, val):
        if val.lower()[:3] in self._functions:
            self._shape = val.lower()
            self._function = self._functions[val.lower()[:3]]
        else:
            raise KeyError('Unknown function shape `{}`'.format(val))

    @property
    def repetitions_done(self):
        return self._phase / (2 * np.pi)


class NoiseGenerator(Generator):
    """Generates colored noise.

    Implementation of `Generator` for random noise signals.

    Arguments:
        color (`str`, optional): The color of the noise. Each color corresponds
            to a inverse frequency power in the noise power density spectrum.
            Default is ``'white'``.

            - ``'purple'``: -2
            - ``'blue'``: -1
            - ``'white'``: 0
            - ``'pink'``: 1
            - ``'brown'``: 2
        method (`str`): The method used to create the noise.
            Currently two methods are implemented, a ``'fft'`` method and
            an ``'autoregressive'`` method. The default is ``'autoregressive'``.
            The autoregressive method is more expensive for small framesizes,
            but gives the same performance regardless of the framesize. The
            fft method have bad low-frequency performance for small framesizes.
        ar_order (`int`): The order for the autoregressive method, default 63.
    References:
        N. J. Kasdin, “Discrete simulation of colored noise and stochastic
        processes and 1/f^α power law noise generation,” Proceedings of the
        IEEE, vol. 83, no. 5, pp. 802–827, May 1995.
        :doi:`10.1109/5.381848`
    Todo:
        Variable amplitudes and maximum amplitudes.

    """
    _color_slopes = {
        'purple': -2,
        'blue': -1,
        'white': 0,
        'pink': 1,
        'brown': 2
    }

    def __init__(self, color='white', method='autoregressive',
                 ar_order=63, **kwargs):
        Generator.__init__(self, **kwargs)
        self.method = method
        self.ar_order = 63
        self.color = color

    def _fft_noise(self):
        normal = np.random.normal(size=self.device.framesize)
        shaped = ifft(self._spectral_coefficients * fft(normal))
        return shaped

    def _fft_setup(self):
        bins = np.arange(self.device.framesize // 2 + 1)  # We are using single sided spectrum
        bins[0] = 1  # Do not modify DC bin
        self._spectral_coefficients = bins.astype('double')**(-self.power / 2)

    def _fft_reset(self):
        del self._spectral_coefficients

    def _ar_noise(self):
        normal = np.random.normal(size=self.device.framesize)
        shaped = np.zeros(shape=self.device.framesize)
        for idx in range(self.device.framesize):
            shaped[idx] = normal[idx] - (self._ar_coefficients * self._ar_buffer).sum()
            self._ar_buffer = np.roll(self._ar_buffer, 1)
            self._ar_buffer[0] = shaped[idx]
        return shaped

    def _ar_setup(self):
        self._ar_buffer = np.zeros(self.ar_order - 1)
        coefficients = np.zeros(self.ar_order)
        coefficients[0] = 1
        for k in range(1, self.ar_order):
            coefficients[k] = (k - 1 - self.power / 2) * coefficients[k - 1] / k
        self._ar_coefficients = coefficients[1:]

    def _ar_reset(self):
        del self._ar_buffer
        del self._ar_coefficients

    _call_methods = {
        'fft': _fft_noise,
        'autoregressive': _ar_noise
    }

    _setup_methods = {
        'fft': _fft_setup,
        'autoregressive': _ar_setup
    }

    _reset_methods = {
        'fft': _fft_reset,
        'autoregressive': _ar_reset
    }

    def frame(self):
        return self._call_methods[self.method](self)

    def setup(self):
        Generator.setup(self)
        self._setup_methods[self.method](self)

    def reset(self):
        Generator.reset(self)
        self._reset_methods[self.method](self)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, val):
        try:
            self.power = self._color_slopes[val.lower()]
        except KeyError:
            raise KeyError('Unknown noise color `{}`'.format(val))
        except AttributeError:
            self.power = val
            self._color = 'custom'
        else:
            self._color = val.lower()

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, val):
        if val.lower() in self._call_methods:
            self._method = val.lower()
        else:
            raise KeyError('Unknown generation method `{}`'.format(val))



import numpy as np
import queue


def flush_Q(q):
    while True:
        try:
            q.get(timeout=0.1)
        except queue.Empty:
            break


def concatenate_Q(q):
    data_list = []
    while True:
        try:
            data_list.append(q.get(timeout=0.1))
        except queue.Empty:
            break
    return np.concatenate(data_list, axis=-1)



import logging

__version__ = '0.0.2'
logger = logging.getLogger(__name__)

__all__ = ['core',
           'devices',
           'triggers',
           'generators',
           'processors',
           'distributors',
           ]

from . import *


import queue
import numpy as np
import sounddevice as sd
from . import core

try:
    import nidaqmx
    import nidaqmx.stream_readers
    import nidaqmx.stream_writers
except ImportError:
    pass


class AudioDevice(core.Device):
    """Class for interacting with audio interfaces.

    Implementation of the `~.core.Device` framework for audio interfaces.
    Built on top of the `sounddevice <http://python-sounddevice.readthedocs.io/>`_ package.

    Arguments:
        name (`str`): Partial or full name of the audio interface.
        fs (`float`, optional): Sample rate for the device, defaults to system default for the device.
        framesize (`int`, optional): The framesize for inputs and outputs, defaults to 1024 samples.
    """
    @staticmethod
    def get_devices(name=None):
        """Check which audio interfaces can be interacted with

        Arguments:
            name (`str`, optional): incomplete name of device or ``None``
        Returns:
            Complete name of device, or list of all devices.
        Todo:
            - The framesize by default set to 1024, is this appropriate?
            - `sounddevice` allows for variable framesizes, can we make use of this somehow?
        """
        if name is None:
            return sd.query_devices()
        else:
            return sd.query_devices(name)['name']

    def __init__(self, name=None, fs=None, framesize=None):
        core.Device.__init__(self)
        if name is None:
            self.name = AudioDevice.get_devices()[0]['name']
        else:
            self.name = AudioDevice.get_devices(name)

        if fs is None:
            self.fs = sd.query_devices(self.name)['default_samplerate']
        else:
            self.fs = fs

        if framesize is None:
            self.framesize = 1024
        else:
            self.framesize = framesize

        self.inputs = []
        self.outputs = []

    def _hardware_run(self):
        while sd._initialized > 0:
            sd._terminate()
        sd._initialize()
        # Create streams
        num_output_ch = max(self.outputs, default=-1) + 1
        num_input_ch = max(self.inputs, default=-1) + 1
        self.silent_ch = [ch for ch in range(num_output_ch) if ch not in self.outputs]
        self._hardware_output_timeout = 0.5 * self.framesize / self.fs
        if num_input_ch and num_output_ch:
            # in/out stream
            def callback(indata, outdata, frames, time, status):
                self._input_callback(indata)
                self._output_callback(outdata)
            stream = sd.Stream(device=self.name, samplerate=self.fs, blocksize=self.framesize, callback=callback, channels=(num_input_ch, num_output_ch))
        elif num_input_ch:
            def callback(indata, frames, time, status):
                self._input_callback(indata)
            stream = sd.InputStream(device=self.name, samplerate=self.fs, blocksize=self.framesize, callback=callback, channels=num_input_ch)
        elif len(self.outputs):
            def callback(outdata, frames, time, status):
                self._output_callback(outdata)
            stream = sd.OutputStream(device=self.name, samplerate=self.fs, blocksize=self.framesize, callback=callback, channels=num_output_ch)
        else:
            # Do something??
            pass

        stream.start()
        self._hardware_stop_event.wait()
        stream.stop()

    def _input_callback(self, indata):
        self._hardware_input_Q.put(indata.T[self.inputs].copy())  # Transpose before copy to actually reverse the memory layout

    def _output_callback(self, outdata):
        try:
            outdata[:, self.outputs] = self._hardware_output_Q.get(timeout=self._hardware_output_timeout).T
            outdata[:, self.silent_ch] = 0
        except queue.Empty:
            outdata[:] = 0

    @property
    def max_inputs(self):
        return sd.query_devices(self.name)['max_input_channels']

    @property
    def max_outputs(self):
        return sd.query_devices(self.name)['max_output_channels']


class NIDevice(core.Device):
    """Class for interacting with national instruments hardware.

    Implementation of the `~.core.Device` framework for national instruments hardware.
    Built on top of the `nidaqmx <https://nidaqmx-python.readthedocs.io/>`_ package.

    Arguments:
        name (`str`): Partial or full name of the audio interface.
        fs (`float`, optional): Sample rate for the device, defaults maximum supported rate.
        framesize (`int`, optional): The framesize for inputs and outputs, defaults to 10000 samples.
        dtype (`str`, optional): The datatype used while reading input data, default ``'float64'``.
    Todo:
        The framesize needs reasonable defaults, and needs to be protected
        by validity checks.
    """
    @staticmethod
    def get_devices(name=None):
        """Check which NI hardware is available.

        Since `NIDevice` is intended to interact with a single module at the
        time, this will complete names by selecting the first available module
        is a chassi.

        Arguments:
            name (`str`, optional): incomplete name of device or ``None``
        Returns:
            Complete name of device, or list of all devices.
        """
        try:
            system = nidaqmx.system.System.local()
        except NameError as e:
            if e.args[0] == "name 'nidaqmx' is not defined":
                raise ModuleNotFoundError("Windows-only module 'nidaqmx' is not installed")
            else:
                raise e
        name_list = [dev.name for dev in system.devices]
        if name is None:
            return name_list
        else:
            if len(name) == 0:
                name = name_list[0]
            if name[:4] == 'cDAQ' and name[5:8] != 'Mod':
                name = [x for x in name_list if x[:8] == name[:5] + 'Mod'][0]
            return name

    def __init__(self, name=None, fs=None, framesize=10000, dtype='float64'):
        core.Device.__init__(self)
        if name is None:
            name = NIDevice.get_devices()[0]
        self.name = NIDevice.get_devices(name)
        if fs is None:
            try:
                self.fs = nidaqmx.system.Device(self.name).ai_max_single_chan_rate
            except nidaqmx.DaqError as e:
                if e.error_type == nidaqmx.error_codes.DAQmxErrors.ATTR_NOT_SUPPORTED:
                    self.fs = nidaqmx.system.Device(self.name).ao_max_rate
                else:
                    raise
        else:
            self.fs = fs
        self.framesize = framesize
        self.dtype = dtype

    @property
    def max_inputs(self):
        return len(nidaqmx.system.Device(self.name).ai_physical_chans)

    @property
    def max_outputs(self):
        return len(nidaqmx.system.Device(self.name).ao_physical_chans)

    @property
    def input_range(self):
        """Returns the input range on the device.

        Note:
            This is only an approximate value, do NOT use for calibrating
            unscaled readings.
        """
        return nidaqmx.system.Device(self.name).ai_voltage_rngs

    @property
    def output_range(self):
        """Returns the output range on the device.

        Note:
            This is only an approximate value, do NOT use for calibrating
            unscaled outputs.
        """
        return nidaqmx.system.Device(self.name).ao_voltage_rngs

    def bit_depth(self, channel=None):
        """The bitdepth for the device.

        Currently only implemented for input devices.

        Todo:
            - What would be the expected behavior if the channels have different depths?
            - Chack that the task exists, else warn.
        """
        if channel is None:
            ch_idx = 0
        else:
            ch_idx = self.inputs.index(channel)
        return self._task.ai_channels[ch_idx].ai_resolution

    def word_length(self, channel=None):
        """The word length for the device.

        Only valid when using raw, unscaled, data types.
        Currently only implemented for input devices.

        Todo:
            - Chack that the task exists, else warn.
        """
        if channel is None:
            return max([ch.ai_raw_samp_size for ch in self._task.ai_channels])
        else:
            ch_idx = self.inputs.index(channel)
            return self._task.ai_channels[ch_idx].ai_raw_samp_size

    def scaling_coeffs(self, channels=None):
        """Scaling coefficients used while reading raw input.

        Returns the polynomial coefficients required to calculate
        input voltage from unscaled integers.

        Todo:
            - Chack that the task exists, else warn.
        """
        if channels is None:
            channels = self.inputs
        try:
            ch_idx = [self.inputs.index(ch) for ch in channels]
        except TypeError:
            ch_idx = self.inputs.index(channels)
            return self._task.ai_channels[ch_idx].ai_dev_scaling_coeff
        else:
            return [self._task.ai_channels[idx].ai_dev_scaling_coeff for idx in ch_idx]

    def _hardware_run(self):
        """
        Todo:
            - Merge the callback creation function internally in this function.
                Since they will never be used outside of here, and only once,
                it does not make much sense to keep them as functions.
        """
        self._task = nidaqmx.Task()
        for ch in self.inputs:
            self._task.ai_channels.add_ai_voltage_chan(self.name + '/ai{}'.format(int(ch)))
        for ch in self.outputs:
            self._task.ao_channels.add_ao_voltage_chan(self.name + '/ao{}'.format(int(ch)), min_val=self.output_range[0], max_val=self.output_range[1])
        if len(self.inputs) or len(self.outputs):
            self._task.timing.cfg_samp_clk_timing(int(self.fs), samps_per_chan=self.framesize,
                    sample_mode=nidaqmx.constants.AcquisitionType.CONTINUOUS)

        if self.dtype.lower() == 'unscaled' or self.dtype.lower() == 'int':
            wl = self.word_length()
            self.dtype = 'int{}'.format(int(wl))
        if len(self.inputs) > 0:
            self._task.register_every_n_samples_acquired_into_buffer_event(self.framesize, self._create_input_callback())
        if len(self.outputs):
                self._task.register_every_n_samples_transferred_from_buffer_event(self.framesize, self._create_output_callback())

        self._task.start()
        self._hardware_stop_event.wait()
        self._task.stop()
        self._task.wait_until_done(timeout=10)
        self._task.close()
        del self._task
        nidaqmx.system.Device(self.name).reset_device()

    def _create_input_callback(self):
        if self.dtype == 'int16':
            reader = nidaqmx.stream_readers.AnalogUnscaledReader(self._task.in_stream)
            databuffer = np.empty((len(self.inputs), self.framesize), dtype='int16')
            read_function = reader.read_int16
        elif self.dtype == 'int32':
            reader = nidaqmx.stream_readers.AnalogUnscaledReader(self._task.in_stream)
            databuffer = np.empty((len(self.inputs), self.framesize), dtype='int32')
            read_function = reader.read_int32
        elif self.dtype == 'uint16':
            reader = nidaqmx.stream_readers.AnalogUnscaledReader(self._task.in_stream)
            databuffer = np.empty((len(self.inputs), self.framesize), dtype='uint16')
            read_function = reader.read_uint16
        elif self.dtype == 'uint32':
            reader = nidaqmx.stream_readers.AnalogUnscaledReader(self._task.in_stream)
            databuffer = np.empty((len(self.inputs), self.framesize), dtype='uint32')
            read_function = reader.read_uint32
        else:  # Read as scaled float64
            reader = nidaqmx.stream_readers.AnalogMultiChannelReader(self._task.in_stream)
            databuffer = np.empty((len(self.inputs), self.framesize), dtype='float64')
            read_function = reader.read_many_sample

        def input_callback(task_handle, every_n_samples_event_type,
                           number_of_samples, callback_data):
            sampsRead = read_function(databuffer, self.framesize)
            self._hardware_input_Q.put(databuffer.copy())
            return 0
        return input_callback

    def _create_output_callback(self):
        writer = nidaqmx.stream_writers.AnalogMultiChannelWriter(self._task.out_stream)
        write_funciton = writer.write_many_sample
        self._task.out_stream.regen_mode = nidaqmx.constants.RegenerationMode.DONT_ALLOW_REGENERATION  # Needed to prevent issues with buffer overwrites and reuse
        write_funciton(np.zeros((self._task.out_stream.num_chans, 2*self.framesize)))  # Pre-fill the buffer with zeros, there needs to be something in the buffer when we start
        timeout = 0.5 * self.framesize / self.fs

        def output_callback(task_handle, every_n_samples_event_type,
                            number_of_samples, callback_data):
            try:
                data = self._hardware_output_Q.get(timeout=timeout)
            except queue.Empty:
                data = np.zeros((self._task.out_stream.num_chans, number_of_samples))
            sampsWritten = write_funciton(data)
            return 0
        return output_callback



import os.path
import h5py
import numpy as np
import multiprocessing
import threading
import queue


class Distributor:
    """Base class for Distributors.

    A `Distributor` is an object that should receive the input data from a
    Device, e.g. a plotter or a file writer. Refer to specific implementations
    for more details.
    """
    def __init__(self, device=None, **kwargs):
        self.device = device
        for key, value in kwargs.items():
            setattr(self, key, value)

    def reset(self):
        """Resets the distributor"""
        pass

    def setup(self):
        """Configures the distributor state"""
        pass

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, dev):
        self._device = dev


class HDFWriter(Distributor):
    """ Implements writing to an HDF 5 file.

    Hiearchial Data Format (HDF) 5 is a format suitable for multidimentional
    data. The format supports arbritrary metadata tags, and datasets can be
    organized in a folder-like structure within a single file.

    Fileaccess is restricted to a single writer to maintain file integrity.
    A single writer can be be used to write data with multiple devices at the
    same time by using different datasets.

    HDFWriter supports three different file writing modes, see `start`.

    Arguments:
        name (`str`): The path to the file to write to.
    Note:
        Only create a single HDFWriter per file!
    """
    _timeout = 0.1

    def __init__(self, filename=None):
        if filename is None:
            filename = 'data'  # Default filename

        name, ext = os.path.splitext(filename)
        if len(ext) < 2:  # Either empty or just a dot
            ext = '.h5'
            filename = name + ext
        self.filename = filename
        self._file = None
        self._group = None
        self._devices = []
        self._input_Qs = []
        self._internal_Q = multiprocessing.Queue()
        self._datasets = []
        self._manual_dataset = None
        self._started = False

        self._stop_event = threading.Event()

    @property
    def device(self):
        return self._devices

    @device.setter
    def device(self, device):
        self.add_input(device=device)

    def add_input(self, device=None, Q=None):
        """Adds a new device or queue.

        At least one of ``device`` or ``Q`` must be given. If ``device`` is
        ``None`` or a string the queue will be treated as an deviceless queue.
        If ``Q`` is not given it will be created and registred to the device.

        Arguments:
            device: A device which reads data.
            Q: A queue with data to write.
        """
        if device is None and Q is None:
            raise ValueError('Either `device` or `Q` must be given as input')
        if Q is None:
            Q = device._register_input_Q()
        self._devices.append(device)
        self._input_Qs.append(Q)

    def setup(self):
        if not self._started:
            self.start()

    def start(self, mode='auto', use_process=True):
        """Starts the file writer.

        The file writer will be started in one of three modes with different
        level of user controll.

        Mode: ``'auto'``
            The auto mode doc
        Mode: ``'signal'``
            The signal mode doc
        Mode: ``'manual'``
            The manual mode doc

        Arguments:
            mode (`str`): The mode to start in.
            use_process (`bool`): If writing should be managed in a separate
                process. This is recommended, and enabled by default, since
                file access with `h5py` will block all other threads. If set
                to false, file writing will instead be managed in a thread.
        """
        self.mode = mode
        if use_process:
            self._process = multiprocessing.Process(target=self._write_target)
        else:
            self._process = threading.Thread(target=self._write_target)
        self._process.start()
        if mode == 'auto':
            self._thread = threading.Thread(target=self._auto_target)
            self._thread.start()
        elif mode == 'signal':
            self._write_signal = threading.Event()
            self._thread = threading.Thread(target=self._signal_target)
            self._thread.start()
        elif mode == 'manual':
            class Dummy_thread:
                def join():
                    return True
            self._thread = Dummy_thread
        self._started = True

    def stop(self):
        self._stop_event.set()
        self._thread.join()
        self.write_device_configs()
        self._internal_Q.put(None)
        self._stop_event.clear()

    def select_group(self, **kwargs):
        self._internal_Q.put(('select', kwargs))

    def _select_group(self, group=None):
        if group is None:
            group = '/'
        self._group = self._file.require_group(group)

    def create_dataset(self, **kwargs):
        self._internal_Q.put(('create', kwargs))

    def _create_dataset(self, name=None, ndim=2, index=None, **kwargs):
        if name is None:
            name = 'data{}'.format(index).replace('None', '')  # Default name for sets
        if name in self._group:
            name_idx = 0
            while name + '_' + str(name_idx) in self._group:
                name_idx += 1
            name = name + '_' + str(name_idx)
        kwargs.setdefault('shape', ndim * (1,))
        kwargs.setdefault('maxshape', ndim * (None,))
        kwargs.setdefault('dtype', 'float64')
        dataset = [self._group.create_dataset(name=name, **kwargs), np.array(ndim * (0,)), None]
        if index is None:
            self._manual_dataset = dataset
        elif len(self._datasets) > index:
            self._datasets[index] = dataset
        else:
            skips = index - len(self._datasets)
            self._datasets.extend(skips * [None])
            self._datasets.append(dataset)
        return dataset

    def write(self, **kwargs):
        """
        Todo:
            - Debug the ``step`` keyword. We want it to work both for True/False
                in both manual writes and signal mode, as well as with an index
                or a list of indices for the signalled mode.
            - Allow for manual writes regardless of the mode. If the write function
                is called with some data or a Q, write that data to the manual dataset.
        """
        if self.mode == 'auto':
            pass
        elif self.mode == 'signal':
            if 'step' in kwargs:
                for idx in range(len(self._input_Qs)):
                    self.step(axis=kwargs['step'], index=idx)
            self._write_signal.set()
        elif self.mode == 'manual':

            idx = kwargs.pop('index', None)
            if 'name' in kwargs:
                create_args = kwargs.copy()
                [create_args.pop(key, None) for key in ['Q', 'data', 'step']]
                self.create_dataset(create_args)
            if 'Q' in kwargs:
                while True:
                    try:
                        self._internal_Q.put(('write', (kwargs['Q'].get(timeout=self._timeout), idx)))
                    except queue.Empty:
                        break
            elif 'data' in kwargs:
                self._internal_Q.put(('write', (kwargs['data'], idx)))
            if 'step' in kwargs:
                self.step(axis=kwargs['step'])

    def _write_attrs(self, index=None, **kwargs):
        if index is None:
            attrs = self._group.attrs
        else:
            attrs = self._datasets[index].attrs
        for key, value in kwargs.items():
            attrs[key] = value

    def write_attrs(self, **kwargs):
        self._internal_Q.put(('attrs', kwargs))

    def write_device_configs(self, index=None, *args, **kwargs):
        if index is None:
            for idx in range(len(self._devices)):
                self.write_device_configs(index=idx, *args, **kwargs)
        else:
            attr_names = {'fs', 'label', 'name', 'serial_number'}
            device = self._devices[index]
            # Qs without devices will correspond to None here, so attrs will be an empty dict => only kwargs will be written
            attrs = {}
            for attr in attr_names.union(args):
                try:
                    value = getattr(device, attr)
                except AttributeError:
                    pass
                else:
                    attrs[attr] = value
            try:
                inputs = device.inputs
            except AttributeError:
                pass
            else:
                attrs['input_channels'] = np.string_([ch.to_json() for ch in inputs])
            try:
                outputs = device.outputs
            except AttributeError:
                pass
            else:
                attrs['output_channels'] = np.string_([ch.to_json() for ch in outputs])
            attrs.update(kwargs)
            self.write_attrs(index=index, **attrs)

    def _write(self, data, index=None):
        if index is None:
            # Not an automated write from device Q
            try:
                dataset, head, _ = self._manual_dataset
            except TypeError:
                # We get `TypeError` if `manual_datast` is `None`
                dataset, head, _ = self._create_dataset(ndim=data.ndim, chunks=data.shape)
            self._manual_dataset[2] = data.shape
        else:
            try:
                dataset, head, _ = self._datasets[index]
            except (IndexError, TypeError):
                # We will get an IndexError if index indicated a hogher number dataset than what already exists,
                # but if we have a sparse creation of sets, e.g. set 0 is missing, but set 1 exists, we will get
                # a type error since `None` is not iterable
                dataset, head, _ = self._create_dataset(ndim=data.ndim, chunks=data.shape, index=index)
            self._datasets[index][2] = data.shape

        for idx in range(data.ndim):
            ax = head.size - data.ndim + idx
            if head[ax] + data.shape[idx] > dataset.shape[ax]:
                dataset.resize(head[ax] + data.shape[idx], axis=ax)

        # All indices exept the last ndim number are constant
        idx_list = list(head[:-data.ndim])
        # The last indices should be sliced from head to head+data.shape
        idx_list.extend([slice(start, start + length) for start, length in zip(head[-data.ndim:], data.shape)])
        # The list must be converted to a tuple
        dataset[tuple(idx_list)] = data
        # Uptade the head
        head[-1] += data.shape[-1]

    def step(self, **kwargs):
        self._internal_Q.put(('step', kwargs))

    def _step(self, axis=True, index=None):
        """
        Todo:
            Give a warning if someone tries to step a non-existing dataset
        """
        if index is None:
            try:
                dataset, head, data_shape = self._manual_dataset
            except TypeError:
                # This happens if someone tries to step before datasets are created
                return
        else:
            try:
                dataset, head, data_shape = self._datasets[index]
            except TypeError:
                # TypeError if someone tries to step before datasets are created
                return

        if isinstance(axis, bool):
            axis = max(len(head) - len(data_shape) - 1, 0)
        if axis < 0:
            axis = len(head) + axis - 1
        steps = (len(head) - len(data_shape)) * (1,) + data_shape
        head[axis] += steps[axis]
        head[axis + 1:] = 0
        if axis < len(head) - len(data_shape) and head[axis] >= dataset.shape[axis]:
            dataset.resize(head[axis] + 1, axis)

        return
        # Old implementation below
        if isinstance(axis, bool):
            # Leave the data dimentions intact, step along the next one
            axis = -len(data_shape) - 1
        if axis >= 0:
            # We would like to always index from the rear since the dimentions align there
            axis = axis - dataset.ndim

        if -axis <= len(data_shape):
            # Step along axis in data
            # Reshaping as a consequence of this cannot be done here since we allow a new data shape
            head[axis] += data_shape[axis]
        else:
            # Step along axis not existing in data, resize if we need to
            head[axis] += 1
            if head[axis] >= dataset.shape[axis]:
                dataset.resize(head[axis] + 1, dataset.ndim + axis)
        # Reset axes after step axis
        # Strange indexing needed for when axis=-1, since there is no -0 equivalent
        head[head.size + axis + 1:] = 0
        # self.head = head

    def _write_target(self):
        def handle(item):
            if item is None:
                raise StopIteration
            else:
                action = item[0]
                args = item[1]
            if action == 'write':
                self._write(*args)
            elif action == 'create':
                dataset = self._create_dataset(**args)
            elif action == 'step':
                self._step(**args)
            elif action == 'select':
                self._select_group(**args)

        self._file = h5py.File(self.filename, mode='a')
        self._select_group()
        # Main write loop
        while True:
            try:
                handle(self._internal_Q.get(timeout=self._timeout))
            except queue.Empty:
                continue
            except StopIteration:
                break
        self._file.close()

    def _auto_target(self):
        for idx, device in enumerate(self._devices):
            name = getattr(device, 'label', getattr(device, 'name', 'dataset{}'.format(idx)))
            self.create_dataset(name=name, ndim=2, index=idx)
        while not self._stop_event.is_set():
            for idx, Q in enumerate(self._input_Qs):
                try:
                    data = Q.get(timeout=self._timeout)
                except queue.Empty:
                    continue
                self._internal_Q.put(('write', (data, idx)))
        # Stop event have bel set, clear out any remaining stuff in th Q
        for idx, Q in enumerate(self._input_Qs):
            while True:
                try:
                    data = Q.get(timeout=self._timeout)
                except queue.Empty:
                    break
                self._internal_Q.put(('write', (data, idx)))

    def _signal_target(self):
        for idx, device in enumerate(self._devices):
            name = getattr(device, 'label', getattr(device, 'name', 'dataset{}'.format(idx)))
            self.create_dataset(name=name, ndim=3, index=idx)
        while not self._stop_event.is_set():
            if self._write_signal.wait(self._timeout):
                self._write_signal.clear()
                for Q in self._input_Qs:
                    # We should acuire all the locks as fast a possible!
                    Q.mutex.acquire()
                for idx, Q in enumerate(self._input_Qs):
                    for data in Q.queue:
                        self._internal_Q.put(('write', (data, idx)))
                    Q.queue.clear()
                for Q in self._input_Qs:
                    Q.mutex.release()


class HDFReader:
    """ Implements reading from HDF 5 files"""

    def __init__(self, filename):
        if not os.path.isfile(filename):
            raise IOError('No such file: {}'.format(filename))
        self.file = h5py.File(filename, mode='r')
        self.datasets = []
        for dataset in self.file:
            self.datasets.append(dataset)
        self.dataset = self.datasets[0]

    def __del__(self):
        self.close()

    def close(self):
        if self.file.id:
            self.file.close()

    @property
    def shape(self):
        return self.dataset.shape

    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        if dataset not in self.datasets:
            if isinstance(dataset, int):
                dataset = self.datasets[dataset]
            else:
                raise KeyError(dataset)
        self._dataset = self.file[dataset]

    def blocks(self, start=None, stop=None, blocksize=None):
        """ A block generator. The last block might be of a different shape than the rest."""
        if start is None:
            start = 0
        if stop is None:
            stop = self.shape[1]
        if start < 0:
            start = self.shape[1] + start
        if stop < 0:
            stop = self.shape[1] + stop
        if blocksize is None:
            blocksize = self.dataset.chunks[1]
        n_blocks = -((start - stop) // blocksize)  # Round up
        for block in range(n_blocks):
            yield self[:, start + block * blocksize:min(start + (block + 1) * blocksize, stop)]

    def __getitem__(self, key):
        if len(key) > self.dataset.ndim:
            self.dataset = (key[-1])
            return self.dataset[key[:-1]]
        else:
            return self.dataset[key]



import queue
import threading
# import multiprocessing
import collections
import numpy as np
from . import utils
import json
from .generators import GeneratorStop


class Device:
    """Abstract class that provides a consistent framework for different hardware.

    An instance of a specific implementation of of `Device` is typically linked
    to a single physical input/output device. The instance manages connections to
    the device, any attached triggers, enabling/disabling input/output, and
    attached generators.

    Attributes:
        input_active (`~threading.Event`): Controls input state.
            Use `~threading.Event.set` to activate input and `~threading.Event.clear` to deactivate.
        output_active (`~threading.Event`): Controls output state.
            Use `~threading.Event.set` to activate output and `~threading.Event.clear` to deactivate.
        inputs (list[`Channel`]): List of assigned inputs, see `add_input`.
        outputs (list[`Channel`]): List of assigned outputs, see `add_output`.
        max_inputs (`int`): The maximum number of inputs available.
        max_outputs (`int`): The maximum number of outputs available.
        calibrations (`numpy.ndarray`): Calibrations of input channels, defaults to 1 for missing calibrations.
    """
    _generator_timeout = 1
    _trigger_timeout = 1
    _q_timeout = 1
    _hardware_timeout = 1

    def __init__(self):
        # self.input_active = multiprocessing.Event()
        self.input_active = threading.Event()
        # self.output_active = multiprocessing.Event()
        self.output_active = threading.Event()

        self.inputs = []
        self.outputs = []

        self.__generators = []
        self.__triggers = []
        self.__Qs = []
        self.__distributors = []

        # self.__main_stop_event = multiprocessing.Event()
        self.__main_stop_event = threading.Event()
        # self.__main_thread = multiprocessing.Process()
        self.__main_thread = threading.Thread()

    def start(self):
        """Starts the device.

        This creates the connections between the hardware and the software,
        configures the hardware, and initializes triggers and generators.
        Triggers are activated unless manually deactivated beforehand.
        Generators will not start generating data until the output is activated.

        Note:
            This does NOT activate inputs or outputs!
        """
        # self.__main_thread = multiprocessing.Process(target=self._Device__main_target)
        self.__main_thread = threading.Thread(target=self._Device__main_target)

        self.__main_thread.start()

    def stop(self):
        """Stops the device.

        Use this to turn off or disconnect a device safely after a measurement.
        It is not recommended to use this as deactivation control, i.e. you should
        normally not have to make multiple calls to this function.

        """
        self.__main_stop_event.set()
        # self.__process.join(timeout=10)

    def add_input(self, index, **kwargs):
        """Adds a new input `Channel`.

        Arguments:
            index (`int`): Zero-based index of the channel.
            **kwargs: All arguments of `Channel` except ``chtype`` and ``index``.

        """
        if index not in self.inputs and index < self.max_inputs:
            self.inputs.append(Channel(index, 'input', **kwargs))

    def add_output(self, index, **kwargs):
        """Adds a new output `Channel`.

        Arguments:
            index (`int`): Zero-based index of the channel.
            **kwargs: All arguments of `Channel` except ``chtype`` and ``index``.
        """
        if index not in self.outputs and index < self.max_outputs:
            self.outputs.append(Channel(index, 'output', **kwargs))

    @property
    def calibrations(self):
        return np.array([c.calibration if c.calibration is not None else 1 for c in self.inputs])

    def _input_scaling(self, frame):
        """Scales the input for triggering

        This is separate from applying calibration values, which is controlled
        by each trigger. The intention here is to account for data formats,
        e.g. reading data as int32 instead of floats or removing DC offsets.

        Arguments:
            frame (`numpy.ndarray`): Unscaled input frame.
        Returns:
            `numpy.ndarray`: Scaled input frame
        """
        return frame

    @property
    def max_inputs(self):
        return np.inf

    @property
    def max_outputs(self):
        return np.inf

    def _hardware_run(self):
        """This is the primary method in which hardware interfacing should be implemented.

        This method will run in a separate thread. It is responsible for creating the
        connections to the hardware according to the configurations.
        If the device has registered inputs, they should always be read and frames
        should be put in the input Q (in the order registered).
        If the device has registered outputs, frames should be taken from the output Q
        and and output through the physical channels (in the order registered).
        If a specific hardware requires constant data streams, fill the stream
        with zeros if no output data is available. The timings for this must be
        implemented by this method.

        This method is responsible for checking if the stop event is set with the interval
        specified by the hardware timeout attribute. When this thread receives the signal
        to stop, it must close all connections to the device and reset it to a state from
        where it can be started again (possibly from another instance).

        Attributes:
            _hardware_input_Q (`~queue.Queue`): The Q where read input data should go.
            _hardware_output_Q (`~queue.Queue`): The Q where the output data to write is stored.
            _hardware_stop_event (`~threading.Event`): Tells the hardware thread to stop.
            _hardware_timeout (`float`): How often the stop event should be checked.
        """
        raise NotImplementedError('Required method `_hardware_run` not implemented in {}'.format(self.__class__.__name__))

    def flush(self):
        """Used to flush all Qs.

        This can be useful if a measurement needs to be discarded.
        Data that have been removed from the queues, e.g. automatic file writers,
        will not be interfered with.

        Note:
            This will delete data which is still in the queues!
        """
        for q in self.__Qs:
            utils.flush_Q(q)

    def input_data(self):
        """Collects the acquired input data.

        Data in stored internally in the `Device` object while input is active.
        This method is a convenient way to access the data from a measurement
        when more elaborate and automated setups are not required.

        Returns:
            `numpy.ndarray`: Array with the input data.
            Has the shape (n_inputs, n_samples), and the input channels are
            ordered in the same order as they were added.
        """
        if self.input_active.is_set():
            print('It is not safe to get all data while input is active!')
        else:
            return utils.concatenate_Q(self.__internal_input_Q)

    def calibrate(self, channel, frequency=1e3, value=1, ctype='rms', unit='V'):
        """Calibrates a channel using a reference signal.

        The resulting calibration value and unit is stored as attributes of the
        corresponding `Channel`. Different calibration types should be used for different
        instruments.
        Currently only unfiltered RMS calibrations are implemented. This detects the level
        in the signal for 3 seconds, and uses the final level as the calibration value.

        Arguments:
            channel (`int`): Index of the channel, in the order that they were added to the device.
            frequency (`float`): The frequency of the applied reference signal, defaults to 1 kHz.
            value (`float`): The value of the reference signal, defaults to 1.
            ctype (``'rms'``): Use to switch between different calibration methods. Currently not used.
            unit (`str`): The unit of the calibrated quantity, defaults to ``'V'``.
        Todo:
            - Filtering the input before detecting the level
            - Average over multiple parts
            - Determine a reasonable value of the averaging coefficient
        """
        detector = utils.LevelDetector(channel=channel, fs=self.fs, time_constant=12/frequency)
        timer = threading.Timer(interval=3, function=lambda x: self.__triggers.remove(x), args=(detector,))
        self.__triggers.append(detector)
        timer.start()
        timer.join()
        channel = self.inputs[self.inputs.index(channel)]
        channel.calibration = detector.current_level / value
        channel.unit = unit

    def _register_input_Q(self, Q=None):
        """Registers new input Q.

        This should be used to register a queue used by a Distributor. The
        queue will receive frames read while the input is active.
        For memory efficiency the input frames are not copied to individual
        queues, so in-place operations are not safe. If a Distributor needs
        to manipulate the data a copy should be made before manipulation.

        Arguments:
            Q (`~queue.Queue`, optional): The Q to register. Will be created if equal to `None`
        Returns:
            `~queue.Queue`: The registered Q.
        Note:
            The frames are NOT copied to multiple queues!
        Todo:
            Give a warning instead of an error while running.

        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to register new Qs while the device is running. Stop the device and perform all setup before starting.')
        else:
            # Q = multiprocessing.Queue()
            if Q is None:
                Q = queue.Queue()
            self.__Qs.append(Q)
            return Q

    def _unregister_input_Q(self, Q):
        """Unregisters input Q.

        Removes a queue from the list of queues that receive input data.
        This method should be used by a Distributor if it is removed from
        the Device.

        Arguments:
            Q (`~queue.Queue`): The Q to remove.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to remove Qs while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__Qs.remove(Q)

    def add_distributor(self, distributor):
        """Adds a Distributor to the Device.

        Arguments:
            distributor: The distributor to add.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to add distributors while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__distributors.append(distributor)
            distributor.device = self

    def remove_distributor(self, distributor):
        """Removes a Distributor from the Device.

        Arguments:
            distributor: The distributor to remove.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to remove distributors while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__distributors.remove(distributor)
            try:
                distributor.remove(self)
            except AttributeError:
                distributor.device = None

    def add_trigger(self, trigger):
        """Adds a Trigger to the Device.

        Arguments:
            trigger: The trigger to add.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to add new triggers while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__triggers.append(trigger)
            trigger.device = self

    def remove_trigger(self, trigger):
        """Removes a Trigger from the Device.

        Arguments:
            trigger: The trigger to remove.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to remove triggers while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__triggers.remove(trigger)
            trigger.device = None

    def add_generator(self, generator):
        """Adds a Generator to the Device.

        Arguments:
            generator: The generator to add.
        Note:
            The order that multiple generators are added to a device
            dictates which output channel receives data from which generator.
            The total number of generated channels must match the number of
            output channels.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to add new generators while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__generators.append(generator)
            generator.device = self

    def remove_generator(self, generator):
        """Removes a Generator from the Device.

        Arguments:
            generator: The generator to remove.
        Todo:
            Give a warning instead of an error while running.
        """
        if self.__main_thread.is_alive():
            raise UserWarning('It is not possible to add new generators while the device is running. Stop the device and perform all setup before starting.')
        else:
            self.__generators.remove(generator)
            generator.device = None

    def __reset(self):
        """Resets the `Device`.

        Performs a number of tasks required to restart a device after it has
        been stopped:

        - Clears all stop events
        - Resets all triggers
        - Resets all generators
        - Clears input and output activation

        """
        self.__main_stop_event.clear()
        self.__trigger_stop_event.clear()
        self.__q_stop_event.clear()
        self._hardware_stop_event.clear()
        for trigger in self.__triggers:
            trigger.reset()
        for generator in self.__generators:
            generator.reset()
        self.input_active.clear()
        self.output_active.clear()

    def _Device__main_target(self):
        """Main method for a Device.

        This is the method that is executed when the device is started.
        Four other threads will be started in this method, one for generators,
        one for triggers, one for queue handling, and one for the hardware.
        This method is also responsible for creating most of the queues and
        events that connect the other threads, as well as initializing the
        distributors.
        """
        # The explicit naming of this method is needed on windows for some stange reason.
        # If we rely on the automatic name wrangling for the process target, it will not be found in device subclasses.
        self._hardware_input_Q = queue.Queue()
        self._hardware_output_Q = queue.Queue(maxsize=25)
        self._hardware_stop_event = threading.Event()
        self.__triggered_q = queue.Queue()
        self.__internal_input_Q = queue.Queue()
        self.__Qs.append(self.__internal_input_Q)
        self.__generator_stop_event = threading.Event()
        self.__trigger_stop_event = threading.Event()
        self.__q_stop_event = threading.Event()
        # Start hardware in separate thread
        # Manage triggers in separate thread
        # Manage Qs in separate thread
        generator_thread = threading.Thread(target=self.__generator_target)
        hardware_thread = threading.Thread(target=self._hardware_run)
        trigger_thread = threading.Thread(target=self.__trigger_target)
        q_thread = threading.Thread(target=self.__q_target)

        generator_thread.start()
        hardware_thread.start()
        trigger_thread.start()
        q_thread.start()

        self.__main_stop_event.wait()

        self.__generator_stop_event.set()
        generator_thread.join()
        self._hardware_stop_event.set()
        hardware_thread.join()
        self.__trigger_stop_event.set()
        trigger_thread.join()
        self.__q_stop_event.set()
        q_thread.join()
        self.__reset()

    def __trigger_target(self):
        """Trigger handling method.

        This method will execute as a subthread in the device, responsible
        for managing the attached triggers, and handling input data.

        Todo:
            - Pre-triggering using appropriate values
            - Post-triggering
            - Aligning triggers?
        """
        data_buffer = collections.deque(maxlen=10)
        for trigger in self.__triggers:
            trigger.setup()

        while not self.__trigger_stop_event.is_set():
            # Wait for a frame, if none has arrived within the set timeout, go back and check stop condition
            try:
                this_frame = self._hardware_input_Q.get(timeout=self._trigger_timeout)
            except queue.Empty:
                continue
            # Execute all triggering conditions
            scaled_frame = self._input_scaling(this_frame)
            for trig in self.__triggers:
                trig(scaled_frame)
            # Move the frame to the buffer
            data_buffer.append(this_frame)
            # If the trigger is active, move everything from the data buffer to the triggered Q
            if self.input_active.is_set():
                while len(data_buffer) > 0:
                    self.__triggered_q.put(data_buffer.popleft())

        # The hardware should have stopped by now, analyze all remaining frames.
        while True:
            try:
                this_frame = self._hardware_input_Q.get(timeout=self._trigger_timeout)
            except queue.Empty:
                break
            scaled_frame = self._input_scaling(this_frame)
            for trig in self.__triggers:
                trig(scaled_frame)
            data_buffer.append(this_frame)
            if self.input_active.is_set():
                while len(data_buffer) > 0:
                    self.__triggered_q.put(data_buffer.popleft())

    def __q_target(self):
        """Queue handling method.

        This method will execute as a subthread in the device, responsible
        for moving data from the input (while active) to the queues used by
        Distributors.
        """
        for distributor in self.__distributors:
            distributor.setup()

        while not self.__q_stop_event.is_set():
            # Wait for a frame, if none has arrived within the set timeout, go back and check stop condition
            try:
                this_frame = self.__triggered_q.get(timeout=self._q_timeout)
            except queue.Empty:
                continue
            for Q in self.__Qs:
                # Copy the frame to all output Qs
                Q.put(this_frame)

        # The triggering should have stopped by now, move the remaining frames.
        while True:
            try:
                this_frame = self.__triggered_q.get(timeout=self._q_timeout)
            except queue.Empty:
                break
            for Q in self.__Qs:
                Q.put(this_frame)

    def __generator_target(self):
        """Generator handling method.

        This method will execute as a subthread in the device, responsible
        for managing attached generators, and generating output frames from
        the generators.
        """
        for generator in self.__generators:
            generator.setup()
        use_prev_frame = False
        while not self.__generator_stop_event.is_set():
            if self.output_active.wait(timeout=self._generator_timeout):
                try:
                    if not use_prev_frame:
                        frame = np.concatenate([generator() for generator in self.__generators])
                except GeneratorStop:
                    self.output_active.clear()
                    continue
                try:
                    self._hardware_output_Q.put(frame, timeout=self._generator_timeout)
                except queue.Full:
                    use_prev_frame = True


class Channel(int):
    """Represents a channel of a device.

    Contains information about a physical channel used.

    Arguments:
        index (`int`): Zero-based index of the channel in the device.
        chtype (``'input'`` or ``'output'``): Type of channel.
        label (`str`, optional): User label for identification of the channel.
        calibration (`float`, optional): Manual calibration value.
        unit (`str`, optional): Physical unit of the calibrated channel.
    """
    @classmethod
    def from_json(cls, json_dict):
        """Creates a channel from json representation.

        Arguments:
            json_dict (`str`): json representation of a dictionary containing
                key-value pairs for the arguments of a `Channel`.
        Returns:
            `Channel`: A channel with the given specification.
        """
        return cls(**json.loads(json_dict))

    def to_json(self):
        """Create json representation of this channel.

        Returns:
            `str`: json representation.
        """
        return json.dumps(self.__dict__)

    def __new__(cls, index, *args, **kwargs):
        return super(Channel, cls).__new__(cls, index)

    def __init__(self, index, chtype, label=None, calibration=None, unit=None):
        self.index = index
        self.chtype = chtype
        self.label = label
        self.calibration = calibration
        self.unit = unit

    def __repr__(self):
        return self.__class__.__name__ + '(' + ', '.join(['{}={}'.format(key, value) for key, value in self.__dict__.items()]) + ')'

    def __str__(self):
        label_str = '' if self.label is None else ' "{}"'.format(self.label)
        calib_str = '' if self.calibration is None else ' ({:.4g} {})'.format(self.calibration, self.unit)
        ch_str = '{chtype} channel {index}'.format(chtype=self.chtype, index=self.index).capitalize()
        return '{ch}{label}{calib}'.format(ch=ch_str, label=label_str, calib=calib_str)



from setuptools import setup

__version__ = 'unknown'
for line in open('acoustics_hardware/__init__.py'):
    if line.startswith('__version__'):
        exec(line)
        break


setup(name='acoustics-hardware',
      version=__version__,
      description='Controlling hardware used in acoustic measurement systems',
      long_description=open('README.rst').read(),
      url='https://github.com/AppliedAcousticsChalmers/acoustics-hardware',
      author='Carl Andersson',
      author_email='carl.andersson@chalmers.se',
      license='MIT',
      packages=['acoustics_hardware'],
      install_requires=[
          'numpy',
          'scipy',
          'h5py',
          'sounddevice',
          'nidaqmx;platform_system=="Windows"',
          'pyserial',
          'SchunkMotionProtocol'],
      )



# from threading import Thread, Event
from serial import Serial
import schunk


class SerialGenerator:  # (Thread):
    @staticmethod
    def get_devices(name=None):
        from serial.tools.list_ports import comports
        devs = comports()
        # No input given, return the list of all connected devices
        if name is None:
            return devs
        # No name given, return first connected port
        # This is probably never what you want, but we can do nothing about it here
        if len(name) == 0:
            return devs[0].device
        # Try name as a port name
        for dev in devs:
            if dev.device.lower() == name.lower():
                return dev.device
        # Try name as a serial device name
        for dev in devs:
            if dev.description.lower().find(name.lower()) >= 0:
                return dev.device

    def __init__(self, name, frequency=1e3, amplitude=1, output_unit='rms', shape='sine',
                 sweeps=0, sweep_time=1, sweep_start=100, sweep_stop=1e3, sweep_spacing='log'):
        # Thread.__init__(self)
        self.device = SerialGenerator.get_devices(name)
        self.ser = Serial(port=self.device, timeout=1, dsrdtr=True)
        self._write('system:remote', 'output:load inf')
        self.output_unit = output_unit
        self.amplitude = amplitude
        self.frequency = frequency
        self.shape = shape
        self.sweeps = sweeps
        self.sweep_time = sweep_time
        self.sweep_start = sweep_start
        self.sweep_stop = sweep_stop
        self.sweep_spacing = sweep_spacing
        self._prev_sweep_settings = {
            'time': None,
            'start': None,
            'stop': None,
            'spacing': None
        }

    def __del__(self):
        self.ser.close()

    def _write(self, *commands):
        '''
        Wrapper for writing to connected device
        '''
        for command in commands:
            self.ser.write(bytes(command + '\n', 'UTF-8'))

    def _read(self):
        return self.ser.readline().decode()[:-2]

    def check_errors(self):
        errors = []
        while True:
            self._write('system:error?')
            error = self._read()
            if error == '+0,"No error"':
                break
            else:
                errors.append(error)
        if len(errors) == 0:
            return False
        else:
            return errors

    def _sweep_setup(self):
        if not self.sweep_start == self._prev_sweep_settings['start']:
            self._prev_sweep_settings['start'] = self.sweep_start
            self._write('frequency:start {}'.format(self.sweep_start))
        if not self.sweep_stop == self._prev_sweep_settings['stop']:
            self._prev_sweep_settings['stop'] = self.sweep_stop
            self._write('frequency:stop {}'.format(self.sweep_stop))
        if not self.sweep_time == self._prev_sweep_settings['time']:
            self._prev_sweep_settings['time'] = self.sweep_time
            self._write('sweep:time {}'.format(self.sweep_time))
        if not self.sweep_spacing == self._prev_sweep_settings['spacing']:
            self._prev_sweep_settings['spacing'] = self.sweep_spacing
            self._write('sweep:spacing {}'.format(self.sweep_spacing))

    def off(self):
        self._write('apply:DC DEF, DEF, 0')

    def on(self):
        if self.sweeps == 0:
            self._write('apply:{} {}, {}'.format(self.shape, self.frequency, self.amplitude))
        elif self.sweeps < 0:
            self._sweep_setup()
            self._write(
                'func:shape {}'.format(self.shape),
                'volt {}'.format(self.amplitude),
                'sweep:state on')
        else:
            self._sweep_setup()
            self._write(
                'func:shape {}'.format(self.shape),
                'trigger:source bus',
                'sweep:state on',
                'volt {}'.format(self.amplitude),
                *self.sweeps * ('*TRG',),
                '*WAI')
            self.off()

    @property
    def output_unit(self):
        self._write('voltage:unit?')
        return self._read()

    @output_unit.setter
    def output_unit(self, value):
        if value.lower() == 'rms':
            self._write('voltage:unit VRMS')
        elif value.lower() == 'peak':
            self._write('voltage:unit VPP')

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        if value.lower()[:3] == 'sin':
            self._shape = 'sinusoid'
        elif value.lower()[:3] == 'squ':
            self._shape = 'square'
        elif value.lower()[:3] == 'tri':
            self._shape = 'triangle'
        elif value.lower()[:3] == 'noi':
            self._shape = 'noise'
        elif value.lower()[:3] == 'dc':
            self._shape = 'dc'


class VariSphere:
    def __init__(self, az_port='COM1', el_port='COM2'):
        self.az = schunk.Module(schunk.SerialConnection(
            0x0B, Serial, port=az_port, baudrate=9600, timeout=1))
        self.el = schunk.Module(schunk.SerialConnection(
            0x0B, Serial, port=el_port, baudrate=9600, timeout=1))

    def move(self, az, el):
        self.az.move_pos(az)
        self.el.move_pos(el)

    def move_blocking(self, az, el):
        self.move(az, el)
        self.wait()

    def stop(self):
        self.az.stop()
        self.el.stop()

    def wait(self):
        self.az.wait_until_position_reached()
        self.el.wait_until_position_reached()

    def reset(self):
        self.move(0, 0)
